@@ -33,9 +33,18 @@ pub enum SerialConsistency {
     LocalSerial = 0x0009,
 }
 
+// LegacyConsistency exists, because Scylla may return a SerialConsistency value
+// as Consistency when returning certain error types - the distinction between
+// Consistency and SerialConsistency is not really a thing in CQL.
+#[derive(Debug, Copy, Clone, PartialEq, Eq, PartialOrd, Ord)]
+pub enum LegacyConsistency {
+    Regular(Consistency),
+    Serial(SerialConsistency),
+}
+
 impl Default for Consistency {
     fn default() -> Self {
-        Consistency::Quorum
+        Consistency::LocalQuorum
     }
 }
 
@@ -45,6 +54,21 @@ impl std::fmt::Display for Consistency {
     }
 }
 
+impl std::fmt::Display for SerialConsistency {
+    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
+        write!(f, "{:?}", self)
+    }
+}
+
+impl std::fmt::Display for LegacyConsistency {
+    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
+        match self {
+            Self::Regular(c) => c.fmt(f),
+            Self::Serial(c) => c.fmt(f),
+        }
+    }
+}
+
 impl From<std::num::TryFromIntError> for ParseError {
     fn from(_err: std::num::TryFromIntError) -> Self {
         ParseError::BadData("Integer conversion out of range".to_string())
@@ -137,7 +161,7 @@ pub fn write_short(v: i16, buf: &mut impl BufMut) {
     buf.put_i16(v);
 }
 
-fn read_short_length(buf: &mut &[u8]) -> Result<usize, ParseError> {
+pub(crate) fn read_short_length(buf: &mut &[u8]) -> Result<usize, ParseError> {
     let v = read_short(buf)?;
     let v: usize = v.try_into()?;
     Ok(v)
@@ -177,6 +201,12 @@ pub fn read_bytes<'a>(buf: &mut &'a [u8]) -> Result<&'a [u8], ParseError> {
     Ok(v)
 }
 
+pub fn read_short_bytes<'a>(buf: &mut &'a [u8]) -> Result<&'a [u8], ParseError> {
+    let len = read_short_length(buf)?;
+    let v = read_raw_bytes(len, buf)?;
+    Ok(v)
+}
+
 pub fn write_bytes(v: &[u8], buf: &mut impl BufMut) -> Result<(), ParseError> {
     write_int_length(v.len(), buf)?;
     buf.put_slice(v);
@@ -398,10 +428,16 @@ fn type_uuid() {
     assert_eq!(u, u2);
 }
 
-pub fn read_consistency(buf: &mut &[u8]) -> Result<Consistency, ParseError> {
+pub fn read_consistency(buf: &mut &[u8]) -> Result<LegacyConsistency, ParseError> {
     let raw = read_short(buf)?;
-    let parsed = Consistency::try_from(raw)
-        .map_err(|_| ParseError::BadData(format!("unknown consistency: {}", raw)))?;
+    let parsed = match Consistency::try_from(raw) {
+        Ok(c) => LegacyConsistency::Regular(c),
+        Err(_) => {
+            let parsed_serial = SerialConsistency::try_from(raw)
+                .map_err(|_| ParseError::BadData(format!("unknown consistency: {}", raw)))?;
+            LegacyConsistency::Serial(parsed_serial)
+        }
+    };
     Ok(parsed)
 }
 
@@ -419,7 +455,7 @@ fn type_consistency() {
     let mut buf = Vec::new();
     write_consistency(c, &mut buf);
     let c2 = read_consistency(&mut &*buf).unwrap();
-    assert_eq!(c, c2);
+    assert_eq!(LegacyConsistency::Regular(c), c2);
 
     let c: i16 = 0x1234;
     buf.clear();
