@@ -3,17 +3,19 @@ use futures::{future::RemoteHandle, FutureExt};
 use tokio::io::{split, AsyncRead, AsyncWrite, AsyncWriteExt, BufReader, BufWriter};
 use tokio::net::{TcpSocket, TcpStream};
 use tokio::sync::{mpsc, oneshot};
-use tracing::{error, trace, warn};
+use tokio::time::Instant;
+use tracing::{debug, error, trace, warn};
 use uuid::Uuid;
 
 #[cfg(feature = "ssl")]
 use openssl::ssl::{Ssl, SslContext};
 #[cfg(feature = "ssl")]
 use std::pin::Pin;
+use std::sync::atomic::AtomicU64;
 #[cfg(feature = "ssl")]
 use tokio_openssl::SslStream;
 
-use std::collections::HashMap;
+use std::collections::{BTreeSet, HashMap};
 use std::convert::TryFrom;
 use std::io::ErrorKind;
 use std::net::SocketAddr;
@@ -30,7 +32,7 @@ use crate::batch::{Batch, BatchStatement};
 use crate::frame::{
     self,
     request::{self, batch, execute, query, register, Request},
-    response::{event::Event, result, result::ColumnSpec, Response, ResponseOpcode},
+    response::{event::Event, result, Response, ResponseOpcode},
     server_event_type::EventType,
     value::{BatchValues, ValueList},
     FrameParams, SerializedRequest,
@@ -47,20 +49,81 @@ use crate::transport::Authenticator::{
 };
 use crate::transport::Compression;
 
+// Existing code imports scylla::transport::connection::QueryResult because it used to be located in this file.
+// Reexport QueryResult to avoid breaking the existing code.
+pub use crate::QueryResult;
+
 // Queries for schema agreement
 const LOCAL_VERSION: &str = "SELECT schema_version FROM system.local WHERE key='local'";
 
+// FIXME: Make this constants configurable
+// The term "orphan" refers to stream ids, that were allocated for a {request, response} that no
+// one is waiting anymore (due to cancellation of `Connection::send_request`). Old orphan refers to
+// a stream id, that is orphaned for a long time. This long time is defined below
+// (`OLD_AGE_ORPHAN_THRESHOLD`). Connection, that has a big number (`OLD_ORPHAN_COUNT_THRESHOLD`)
+// of old orphans is shut down (and created again by a connection management layer).
+const OLD_ORPHAN_COUNT_THRESHOLD: usize = 1024;
+const OLD_AGE_ORPHAN_THRESHOLD: std::time::Duration = std::time::Duration::from_secs(1);
+
 pub struct Connection {
     submit_channel: mpsc::Sender<Task>,
     _worker_handle: RemoteHandle<()>,
+
     connect_address: SocketAddr,
-    source_port: u16,
     shard_info: Option<ShardInfo>,
-    config: ConnectionConfig,
     shard_aware_port: Option<u16>,
+    config: ConnectionConfig,
+
+    // Each request send by `Connection::send_request` needs a unique request id.
+    // This field is a monotonic generator of such ids.
+    request_id_generator: AtomicU64,
+    // If a `Connection::send_request` is cancelled, it sends notification
+    // about orphaning via the sender below.
+    // Also, this sender is unbounded, because only unbounded channels support
+    // pushing values in a synchronous way (without an `.await`), which is
+    // needed for pushing values in `Drop` implementations.
+    orphan_notification_sender: mpsc::UnboundedSender<RequestId>,
 }
 
-type ResponseHandler = oneshot::Sender<Result<TaskResponse, QueryError>>;
+type RequestId = u64;
+
+struct ResponseHandler {
+    response_sender: oneshot::Sender<Result<TaskResponse, QueryError>>,
+    request_id: RequestId,
+}
+
+// Used to notify `Connection::orphaner` about `Connection::send_request`
+// future being dropped before receiving response.
+struct OrphanhoodNotifier<'a> {
+    enabled: bool,
+    request_id: RequestId,
+    notification_sender: &'a mpsc::UnboundedSender<RequestId>,
+}
+
+impl<'a> OrphanhoodNotifier<'a> {
+    fn new(
+        request_id: RequestId,
+        notification_sender: &'a mpsc::UnboundedSender<RequestId>,
+    ) -> Self {
+        Self {
+            enabled: true,
+            request_id,
+            notification_sender,
+        }
+    }
+
+    fn disable(mut self) {
+        self.enabled = false;
+    }
+}
+
+impl<'a> Drop for OrphanhoodNotifier<'a> {
+    fn drop(&mut self) {
+        if self.enabled {
+            let _ = self.notification_sender.send(self.request_id);
+        }
+    }
+}
 
 struct Task {
     serialized_request: SerializedRequest,
@@ -79,49 +142,6 @@ pub struct QueryResponse {
     pub warnings: Vec<String>,
 }
 
-/// Result of a single query  
-/// Contains all rows returned by the database and some more information
-#[derive(Default, Debug)]
-pub struct QueryResult {
-    /// Rows returned by the database
-    pub rows: Option<Vec<result::Row>>,
-    /// Warnings returned by the database
-    pub warnings: Vec<String>,
-    /// CQL Tracing uuid - can only be Some if tracing is enabled for this query
-    pub tracing_id: Option<Uuid>,
-    /// Paging state returned from the server
-    pub paging_state: Option<Bytes>,
-    /// Column specification returned from the server
-    pub col_specs: Vec<ColumnSpec>,
-}
-
-impl QueryResult {
-    // Returns a column specification for a column with given name, or None if not found
-    pub fn get_column_spec<'a>(&'a self, name: &str) -> Option<(usize, &'a ColumnSpec)> {
-        self.col_specs
-            .iter()
-            .enumerate()
-            .find(|(_id, spec)| spec.name == name)
-    }
-
-    /// This function is used to merge results of multiple paged queries into one.  
-    /// other is the result of a new paged query.  
-    /// It is merged with current result kept in self.  
-    pub(crate) fn merge_with_next_page_res(&mut self, other: QueryResult) {
-        if let Some(other_rows) = other.rows {
-            match &mut self.rows {
-                Some(self_rows) => self_rows.extend(other_rows),
-                None => self.rows = Some(other_rows),
-            }
-        };
-
-        self.warnings.extend(other.warnings);
-        self.tracing_id = other.tracing_id;
-        self.paging_state = other.paging_state;
-        self.col_specs = other.col_specs;
-    }
-}
-
 /// Result of Session::batch(). Contains no rows, only some useful information.
 pub struct BatchResult {
     /// Warnings returned by the database
@@ -237,25 +257,32 @@ impl Connection {
                 return Err(QueryError::TimeoutError);
             }
         };
-        let source_port = stream.local_addr()?.port();
         stream.set_nodelay(config.tcp_nodelay)?;
 
         // TODO: What should be the size of the channel?
         let (sender, receiver) = mpsc::channel(1024);
-
         let (error_sender, error_receiver) = tokio::sync::oneshot::channel();
+        // Unbounded because it allows for synchronous pushes
+        let (orphan_notification_sender, orphan_notification_receiver) = mpsc::unbounded_channel();
 
-        let _worker_handle =
-            Self::run_router(config.clone(), stream, receiver, error_sender).await?;
+        let _worker_handle = Self::run_router(
+            config.clone(),
+            stream,
+            receiver,
+            error_sender,
+            orphan_notification_receiver,
+        )
+        .await?;
 
         let connection = Connection {
             submit_channel: sender,
             _worker_handle,
-            source_port,
+            config,
             connect_address: addr,
             shard_info: None,
-            config,
             shard_aware_port: None,
+            request_id_generator: AtomicU64::new(0),
+            orphan_notification_sender,
         };
 
         Ok((connection, error_receiver))
@@ -293,6 +320,7 @@ impl Connection {
                 p.prepared_metadata,
                 query.contents.clone(),
                 query.get_page_size(),
+                query.config.clone(),
             ),
             _ => {
                 return Err(QueryError::ProtocolError(
@@ -334,17 +362,6 @@ impl Connection {
         self.query(&query, &values, None).await?.into_query_result()
     }
 
-    pub async fn query_single_page_by_ref(
-        &self,
-        query: &Query,
-        values: &impl ValueList,
-        paging_state: Option<Bytes>,
-    ) -> Result<QueryResult, QueryError> {
-        self.query(query, values, paging_state)
-            .await?
-            .into_query_result()
-    }
-
     pub async fn query(
         &self,
         query: &Query,
@@ -448,7 +465,7 @@ impl Connection {
             .await?;
 
         if let Response::Error(err) = &query_response.response {
-            if err.error == DbError::Unprepared {
+            if let DbError::Unprepared { .. } = err.error {
                 // Repreparation of a statement is needed
                 let reprepare_query: Query = prepared_statement.get_statement().into();
                 let reprepared = self.prepare(&reprepare_query).await?;
@@ -470,6 +487,7 @@ impl Connection {
     }
 
     /// Performs execute_single_page multiple times to fetch all available pages
+    #[allow(dead_code)]
     pub async fn execute_all(
         &self,
         prepared_statement: &PreparedStatement,
@@ -538,19 +556,45 @@ impl Connection {
             timestamp: batch.get_timestamp(),
         };
 
-        let query_response = self
-            .send_request(&batch_frame, true, batch.config.tracing)
-            .await?;
+        loop {
+            let query_response = self
+                .send_request(&batch_frame, true, batch.config.tracing)
+                .await?;
 
-        match query_response.response {
-            Response::Error(err) => Err(err.into()),
-            Response::Result(_) => Ok(BatchResult {
-                warnings: query_response.warnings,
-                tracing_id: query_response.tracing_id,
-            }),
-            _ => Err(QueryError::ProtocolError(
-                "BATCH: Unexpected server response",
-            )),
+            return match query_response.response {
+                Response::Error(err) => match err.error {
+                    DbError::Unprepared { statement_id } => {
+                        let prepared_statement = batch.statements.iter().find_map(|s| match s {
+                            BatchStatement::PreparedStatement(s) if *s.get_id() == statement_id => {
+                                Some(s)
+                            }
+                            _ => None,
+                        });
+                        if let Some(p) = prepared_statement {
+                            let reprepare_query: Query = p.get_statement().into();
+                            let reprepared = self.prepare(&reprepare_query).await?;
+                            if reprepared.get_id() != p.get_id() {
+                                return Err(QueryError::ProtocolError(
+                                    "Prepared statement Id changed, md5 sum should stay the same",
+                                ));
+                            }
+                            continue;
+                        } else {
+                            return Err(QueryError::ProtocolError(
+                                "The server returned a prepared statement Id that did not exist in the batch",
+                            ));
+                        }
+                    }
+                    _ => Err(err.into()),
+                },
+                Response::Result(_) => Ok(BatchResult {
+                    warnings: query_response.warnings,
+                    tracing_id: query_response.tracing_id,
+                }),
+                _ => Err(QueryError::ProtocolError(
+                    "BATCH: Unexpected server response",
+                )),
+            };
         }
     }
 
@@ -620,6 +664,11 @@ impl Connection {
         Ok(version_id)
     }
 
+    fn allocate_request_id(&self) -> RequestId {
+        self.request_id_generator
+            .fetch_add(1, std::sync::atomic::Ordering::Relaxed)
+    }
+
     async fn send_request<R: Request>(
         &self,
         request: &R,
@@ -632,13 +681,23 @@ impl Connection {
             None
         };
         let serialized_request = SerializedRequest::make(request, compression, tracing)?;
+        let request_id = self.allocate_request_id();
 
-        let (sender, receiver) = oneshot::channel();
+        let (response_sender, receiver) = oneshot::channel();
+        let response_handler = ResponseHandler {
+            response_sender,
+            request_id,
+        };
+
+        // Dropping `notifier` (before calling `notifier.disable()`) will send a notification to
+        // `Connection::router`. This notification is then used to mark a `stream_id` associated
+        // with this request as orphaned and free associated resources.
+        let notifier = OrphanhoodNotifier::new(request_id, &self.orphan_notification_sender);
 
         self.submit_channel
             .send(Task {
                 serialized_request,
-                response_handler: sender,
+                response_handler,
             })
             .await
             .map_err(|_| {
@@ -653,9 +712,13 @@ impl Connection {
                 ErrorKind::Other,
                 "Connection broken",
             )))
-        })??;
+        })?;
 
-        Self::parse_response(task_response, self.config.compression)
+        // Response was successfully received, so it's time to disable
+        // notification about orphaning.
+        notifier.disable();
+
+        Self::parse_response(task_response?, self.config.compression)
     }
 
     fn parse_response(
@@ -687,15 +750,28 @@ impl Connection {
         stream: TcpStream,
         receiver: mpsc::Receiver<Task>,
         error_sender: tokio::sync::oneshot::Sender<QueryError>,
+        orphan_notification_receiver: mpsc::UnboundedReceiver<RequestId>,
     ) -> Result<RemoteHandle<()>, std::io::Error> {
         let res = match config.ssl_context {
             Some(ref context) => {
                 let ssl = Ssl::new(context)?;
                 let mut stream = SslStream::new(ssl, stream)?;
                 let _pin = Pin::new(&mut stream).connect().await;
-                Self::run_router_spawner(stream, receiver, error_sender, config)
+                Self::run_router_spawner(
+                    config,
+                    stream,
+                    receiver,
+                    error_sender,
+                    orphan_notification_receiver,
+                )
             }
-            None => Self::run_router_spawner(stream, receiver, error_sender, config),
+            None => Self::run_router_spawner(
+                config,
+                stream,
+                receiver,
+                error_sender,
+                orphan_notification_receiver,
+            ),
         };
         Ok(res)
     }
@@ -706,31 +782,42 @@ impl Connection {
         stream: TcpStream,
         receiver: mpsc::Receiver<Task>,
         error_sender: tokio::sync::oneshot::Sender<QueryError>,
+        orphan_notification_receiver: mpsc::UnboundedReceiver<RequestId>,
     ) -> Result<RemoteHandle<()>, std::io::Error> {
         Ok(Self::run_router_spawner(
+            config,
             stream,
             receiver,
             error_sender,
-            config,
+            orphan_notification_receiver,
         ))
     }
 
     fn run_router_spawner(
+        config: ConnectionConfig,
         stream: (impl AsyncRead + AsyncWrite + Send + 'static),
         receiver: mpsc::Receiver<Task>,
         error_sender: tokio::sync::oneshot::Sender<QueryError>,
-        config: ConnectionConfig,
+        orphan_notification_receiver: mpsc::UnboundedReceiver<RequestId>,
     ) -> RemoteHandle<()> {
-        let (task, handle) = Self::router(stream, receiver, error_sender, config).remote_handle();
+        let (task, handle) = Self::router(
+            config,
+            stream,
+            receiver,
+            error_sender,
+            orphan_notification_receiver,
+        )
+        .remote_handle();
         tokio::task::spawn(task);
         handle
     }
 
     async fn router(
+        config: ConnectionConfig,
         stream: (impl AsyncRead + AsyncWrite),
         receiver: mpsc::Receiver<Task>,
         error_sender: tokio::sync::oneshot::Sender<QueryError>,
-        config: ConnectionConfig,
+        orphan_notification_receiver: mpsc::UnboundedReceiver<RequestId>,
     ) {
         let (read_half, write_half) = split(stream);
         // Why are using a mutex here?
@@ -757,8 +844,9 @@ impl Connection {
             &handler_map,
             receiver,
         );
+        let o = Self::orphaner(&handler_map, orphan_notification_receiver);
 
-        let result = futures::try_join!(r, w);
+        let result = futures::try_join!(r, w, o);
 
         let error: QueryError = match result {
             Ok(_) => return, // Connection was dropped, we can return
@@ -771,7 +859,7 @@ impl Connection {
 
         for (_, handler) in response_handlers {
             // Ignore sending error, request was dropped
-            let _ = handler.send(Err(error.clone()));
+            let _ = handler.response_sender.send(Err(error.clone()));
         }
 
         // If someone is listening for connection errors notify them
@@ -807,24 +895,36 @@ impl Connection {
                 _ => {}
             }
 
-            let handler = {
+            let handler_lookup_res = {
                 // We are guaranteed here that handler_map will not be locked
                 // by anybody else, so we can do try_lock().unwrap()
-                let mut lock = handler_map.try_lock().unwrap();
-                lock.take(params.stream)
+                let mut handler_map_guard = handler_map.try_lock().unwrap();
+                handler_map_guard.lookup(params.stream)
             };
 
-            if let Some(handler) = handler {
-                // Don't care if sending of the response fails. This must
-                // mean that the receiver side was impatient and is not
-                // waiting for the result anymore.
-                let _ = handler.send(Ok(response));
-            } else {
-                // Unsolicited frame. This should not happen and indicates
-                // a bug either in the driver, or in the database
-                return Err(QueryError::ProtocolError(
-                    "Received reponse with unexpected StreamId",
-                ));
+            use HandlerLookupResult::*;
+            match handler_lookup_res {
+                Handler(handler) => {
+                    // Don't care if sending of the response fails. This must
+                    // mean that the receiver side was impatient and is not
+                    // waiting for the result anymore.
+                    let _ = handler.response_sender.send(Ok(response));
+                }
+                Missing => {
+                    // Unsolicited frame. This should not happen and indicates
+                    // a bug either in the driver, or in the database
+                    debug!(
+                        "Received response with unexpected StreamId {}",
+                        params.stream
+                    );
+                    return Err(QueryError::ProtocolError(
+                        "Received reponse with unexpected StreamId",
+                    ));
+                }
+                Orphaned => {
+                    // Do nothing, handler was freed because this stream_id has
+                    // been marked as orphaned
+                }
             }
         }
     }
@@ -835,14 +935,16 @@ impl Connection {
     ) -> Option<i16> {
         // We are guaranteed here that handler_map will not be locked
         // by anybody else, so we can do try_lock().unwrap()
-        let mut lock = handler_map.try_lock().unwrap();
-        if let Some(stream_id) = lock.allocate(response_handler) {
-            Some(stream_id)
-        } else {
-            // TODO: Handle this error better, for now we drop this
-            // request and return an error to the receiver
-            error!("Could not allocate stream id");
-            None
+        let mut handler_map_guard = handler_map.try_lock().unwrap();
+        match handler_map_guard.allocate(response_handler) {
+            Ok(stream_id) => Some(stream_id),
+            Err(response_handler) => {
+                error!("Could not allocate stream id");
+                let _ = response_handler
+                    .response_sender
+                    .send(Err(QueryError::UnableToAllocStreamId));
+                None
+            }
         }
     }
 
@@ -876,6 +978,45 @@ impl Connection {
         Ok(())
     }
 
+    // This task receives notifications from `OrphanhoodNotifier`s and tries to
+    // mark streams as orphaned. It also checks count of old orphans periodically.
+    // After an ald orphan threshold is reached, `orphaner` returns an error
+    // causing the connection to break.
+    async fn orphaner(
+        handler_map: &StdMutex<ResponseHandlerMap>,
+        mut orphan_receiver: mpsc::UnboundedReceiver<RequestId>,
+    ) -> Result<(), QueryError> {
+        let mut interval = tokio::time::interval(OLD_AGE_ORPHAN_THRESHOLD);
+        loop {
+            tokio::select! {
+                _ = interval.tick() => {
+                    // We are guaranteed here that handler_map will not be locked
+                    // by anybody else, so we can do try_lock().unwrap()
+                    let handler_map_guard = handler_map.try_lock().unwrap();
+                    let old_orphan_count = handler_map_guard.old_orphans_count();
+                    if old_orphan_count > OLD_ORPHAN_COUNT_THRESHOLD {
+                        warn!(
+                            "Too many old orphaned stream ids: {}",
+                            old_orphan_count,
+                        );
+                        return Err(QueryError::TooManyOrphanedStreamIds(old_orphan_count as u16))
+                    }
+                }
+                Some(request_id) = orphan_receiver.recv() => {
+                    trace!(
+                        "Trying to orphan stream id associated with request_id = {}",
+                        request_id,
+                    );
+                    let mut handler_map_guard = handler_map.try_lock().unwrap(); // Same as above
+                    handler_map_guard.orphan(request_id);
+                }
+                else => { break }
+            }
+        }
+
+        Ok(())
+    }
+
     async fn handle_event(
         task_response: TaskResponse,
         compression: Option<Compression>,
@@ -902,20 +1043,10 @@ impl Connection {
         &self.shard_info
     }
 
-    /// Are we connected to Scylla's shard aware port?
-    // TODO: couple this with shard_info?
-    pub fn get_is_shard_aware(&self) -> bool {
-        Some(self.connect_address.port()) == self.shard_aware_port
-    }
-
     pub fn get_shard_aware_port(&self) -> Option<u16> {
         self.shard_aware_port
     }
 
-    pub fn get_source_port(&self) -> u16 {
-        self.source_port
-    }
-
     fn set_shard_info(&mut self, shard_info: Option<ShardInfo>) {
         self.shard_info = shard_info
     }
@@ -1083,9 +1214,58 @@ async fn connect_with_source_port(
     }
 }
 
+struct OrphanageTracker {
+    orphans: HashMap<i16, Instant>,
+    by_orphaning_times: BTreeSet<(Instant, i16)>,
+}
+
+impl OrphanageTracker {
+    pub fn new() -> Self {
+        Self {
+            orphans: HashMap::new(),
+            by_orphaning_times: BTreeSet::new(),
+        }
+    }
+
+    pub fn insert(&mut self, stream_id: i16) {
+        let now = Instant::now();
+        self.orphans.insert(stream_id, now);
+        self.by_orphaning_times.insert((now, stream_id));
+    }
+
+    pub fn remove(&mut self, stream_id: i16) {
+        if let Some(time) = self.orphans.remove(&stream_id) {
+            self.by_orphaning_times.remove(&(time, stream_id));
+        }
+    }
+
+    pub fn contains(&self, stream_id: i16) -> bool {
+        self.orphans.contains_key(&stream_id)
+    }
+
+    pub fn orphans_older_than(&self, age: std::time::Duration) -> usize {
+        let minimal_age = Instant::now() - age;
+        self.by_orphaning_times
+            .range(..(minimal_age, i16::MAX))
+            .count() // This has linear time complexity, but in terms of
+                     // the number of old orphans. Healthy connection - one
+                     // that does not have old orphaned stream ids, will
+                     // calculate this function quickly.
+    }
+}
+
 struct ResponseHandlerMap {
     stream_set: StreamIdSet,
     handlers: HashMap<i16, ResponseHandler>,
+
+    request_to_stream: HashMap<RequestId, i16>,
+    orphanage_tracker: OrphanageTracker,
+}
+
+enum HandlerLookupResult {
+    Orphaned,
+    Handler(ResponseHandler),
+    Missing,
 }
 
 impl ResponseHandlerMap {
@@ -1093,19 +1273,63 @@ impl ResponseHandlerMap {
         Self {
             stream_set: StreamIdSet::new(),
             handlers: HashMap::new(),
+            request_to_stream: HashMap::new(),
+            orphanage_tracker: OrphanageTracker::new(),
         }
     }
 
-    pub fn allocate(&mut self, response_handler: ResponseHandler) -> Option<i16> {
-        let stream_id = self.stream_set.allocate()?;
-        let prev_handler = self.handlers.insert(stream_id, response_handler);
-        assert!(prev_handler.is_none());
-        Some(stream_id)
+    pub fn allocate(&mut self, response_handler: ResponseHandler) -> Result<i16, ResponseHandler> {
+        if let Some(stream_id) = self.stream_set.allocate() {
+            self.request_to_stream
+                .insert(response_handler.request_id, stream_id);
+            let prev_handler = self.handlers.insert(stream_id, response_handler);
+            assert!(prev_handler.is_none());
+
+            Ok(stream_id)
+        } else {
+            Err(response_handler)
+        }
     }
 
-    pub fn take(&mut self, stream_id: i16) -> Option<ResponseHandler> {
+    // Orphan stream_id (associated with this request_id) by moving it to
+    // `orphanage_tracker`, and freeing its handler
+    pub fn orphan(&mut self, request_id: RequestId) {
+        if let Some(stream_id) = self.request_to_stream.get(&request_id) {
+            debug!(
+                "Orphaning stream_id = {} associated with request_id = {}",
+                stream_id, request_id
+            );
+            self.orphanage_tracker.insert(*stream_id);
+            self.handlers.remove(stream_id);
+            self.request_to_stream.remove(&request_id);
+        }
+    }
+
+    pub fn old_orphans_count(&self) -> usize {
+        self.orphanage_tracker
+            .orphans_older_than(OLD_AGE_ORPHAN_THRESHOLD)
+    }
+
+    pub fn lookup(&mut self, stream_id: i16) -> HandlerLookupResult {
         self.stream_set.free(stream_id);
-        self.handlers.remove(&stream_id)
+
+        if self.orphanage_tracker.contains(stream_id) {
+            self.orphanage_tracker.remove(stream_id);
+            // This `stream_id` had been orphaned, so its handler got removed.
+            // This is a valid state (as opposed to missing handler)
+            return HandlerLookupResult::Orphaned;
+        }
+
+        if let Some(handler) = self.handlers.remove(&stream_id) {
+            // A mapping `request_id` -> `stream_id` must be removed, to
+            // prevent marking this `stream_id` as orphaned by some late
+            // orphan notification.
+            self.request_to_stream.remove(&handler.request_id);
+
+            HandlerLookupResult::Handler(handler)
+        } else {
+            HandlerLookupResult::Missing
+        }
     }
 
     // Retrieves the map of handlers, used after connection breaks
