@@ -15,10 +15,11 @@ use tokio::sync::mpsc;
 use super::errors::QueryError;
 use crate::cql_to_rust::{FromRow, FromRowError};
 
+use crate::frame::types::LegacyConsistency;
 use crate::frame::{
     response::{
         result,
-        result::{Row, Rows},
+        result::{ColumnSpec, Row, Rows},
         Response,
     },
     value::SerializedValues,
@@ -34,7 +35,7 @@ use crate::transport::node::Node;
 use crate::transport::retry_policy::{QueryInfo, RetryDecision, RetrySession};
 use uuid::Uuid;
 
-/// Iterator over rows returned by paged queries  
+/// Iterator over rows returned by paged queries\
 /// Allows to easily access rows without worrying about handling multiple pages
 pub struct RowIterator {
     current_row_idx: usize,
@@ -59,7 +60,7 @@ pub(crate) struct PreparedIteratorConfig {
     pub metrics: Arc<Metrics>,
 }
 
-/// Fetching pages is asynchronous so `RowIterator` does not implement the `Iterator` trait.  
+/// Fetching pages is asynchronous so `RowIterator` does not implement the `Iterator` trait.\
 /// Instead it uses the asynchronous `Stream` trait
 impl Stream for RowIterator {
     type Item = Result<Row, QueryError>;
@@ -106,7 +107,7 @@ impl RowIterator {
         }
     }
 
-    pub(crate) fn new_for_query(
+    pub(crate) async fn new_for_query(
         query: Query,
         values: SerializedValues,
         default_consistency: Consistency,
@@ -114,8 +115,8 @@ impl RowIterator {
         load_balancer: Arc<dyn LoadBalancingPolicy>,
         cluster_data: Arc<ClusterData>,
         metrics: Arc<Metrics>,
-    ) -> RowIterator {
-        let (sender, receiver) = mpsc::channel(1);
+    ) -> Result<RowIterator, QueryError> {
+        let (sender, mut receiver) = mpsc::channel(1);
         let consistency = query.config.determine_consistency(default_consistency);
 
         let worker_task = async move {
@@ -146,16 +147,24 @@ impl RowIterator {
 
         tokio::task::spawn(worker_task);
 
-        RowIterator {
+        let pages_received = receiver.recv().await.unwrap()?;
+
+        Ok(RowIterator {
             current_row_idx: 0,
-            current_page: Default::default(),
+            current_page: pages_received.rows,
             page_receiver: receiver,
-            tracing_ids: Vec::new(),
-        }
+            tracing_ids: if let Some(tracing_id) = pages_received.tracing_id {
+                vec![tracing_id]
+            } else {
+                Vec::new()
+            },
+        })
     }
 
-    pub(crate) fn new_for_prepared_statement(config: PreparedIteratorConfig) -> RowIterator {
-        let (sender, receiver) = mpsc::channel(1);
+    pub(crate) async fn new_for_prepared_statement(
+        config: PreparedIteratorConfig,
+    ) -> Result<RowIterator, QueryError> {
+        let (sender, mut receiver) = mpsc::channel(1);
         let consistency = config
             .prepared
             .config
@@ -198,12 +207,18 @@ impl RowIterator {
 
         tokio::task::spawn(worker_task);
 
-        RowIterator {
+        let pages_received = receiver.recv().await.unwrap()?;
+
+        Ok(RowIterator {
             current_row_idx: 0,
-            current_page: Default::default(),
+            current_page: pages_received.rows,
             page_receiver: receiver,
-            tracing_ids: Vec::new(),
-        }
+            tracing_ids: if let Some(tracing_id) = pages_received.tracing_id {
+                vec![tracing_id]
+            } else {
+                Vec::new()
+            },
+        })
     }
 
     /// If tracing was enabled returns tracing ids of all finished page queries
@@ -211,6 +226,11 @@ impl RowIterator {
         &self.tracing_ids
     }
 
+    /// Returns specification of row columns
+    pub fn get_column_specs(&self) -> &[ColumnSpec] {
+        &self.current_page.metadata.col_specs
+    }
+
     fn is_current_page_exhausted(&self) -> bool {
         self.current_row_idx >= self.current_page.rows.len()
     }
@@ -278,7 +298,7 @@ where
                 let query_info = QueryInfo {
                     error: &last_error,
                     is_idempotent: self.query_is_idempotent,
-                    consistency: self.query_consistency,
+                    consistency: LegacyConsistency::Regular(self.query_consistency),
                 };
 
                 match self.retry_session.decide_should_retry(query_info) {
@@ -352,7 +372,7 @@ where
 }
 
 /// Iterator over rows returned by paged queries
-/// where each row is parsed as the given type  
+/// where each row is parsed as the given type\
 /// Returned by `RowIterator::into_typed`
 pub struct TypedRowIterator<RowT> {
     row_iterator: RowIterator,
@@ -364,6 +384,11 @@ impl<RowT> TypedRowIterator<RowT> {
     pub fn get_tracing_ids(&self) -> &[Uuid] {
         self.row_iterator.get_tracing_ids()
     }
+
+    /// Returns specification of row columns
+    pub fn get_column_specs(&self) -> &[ColumnSpec] {
+        self.row_iterator.get_column_specs()
+    }
 }
 
 /// Couldn't get next typed row from the iterator
@@ -378,7 +403,7 @@ pub enum NextRowError {
     FromRowError(#[from] FromRowError),
 }
 
-/// Fetching pages is asynchronous so `TypedRowIterator` does not implement the `Iterator` trait.  
+/// Fetching pages is asynchronous so `TypedRowIterator` does not implement the `Iterator` trait.\
 /// Instead it uses the asynchronous `Stream` trait
 impl<RowT: FromRow> Stream for TypedRowIterator<RowT> {
     type Item = Result<RowT, NextRowError>;
