@@ -1,6 +1,7 @@
-//! `Session` is the main object used in the driver.  
+//! `Session` is the main object used in the driver.\
 //! It manages all connections to the cluster and allows to perform queries.
 
+use crate::frame::types::LegacyConsistency;
 use bytes::Bytes;
 use futures::future::join_all;
 use futures::future::try_join_all;
@@ -15,6 +16,7 @@ use uuid::Uuid;
 
 use super::connection::QueryResponse;
 use super::errors::{BadQuery, NewSessionError, QueryError};
+use crate::cql_to_rust::FromRow;
 use crate::frame::response::cql_to_rust::FromRowError;
 use crate::frame::response::result;
 use crate::frame::value::{BatchValues, SerializedValues, ValueList};
@@ -23,24 +25,28 @@ use crate::query::Query;
 use crate::routing::{murmur3_token, Token};
 use crate::statement::{Consistency, SerialConsistency};
 use crate::tracing::{GetTracingConfig, TracingEvent, TracingInfo};
-use crate::transport::connection_pool::PoolConfig;
-use crate::transport::{
-    cluster::Cluster,
-    connection::{BatchResult, Connection, ConnectionConfig, QueryResult, VerifiedKeyspaceName},
-    iterator::RowIterator,
-    load_balancing::{LoadBalancingPolicy, RoundRobinPolicy, Statement, TokenAwarePolicy},
-    metrics::Metrics,
-    node::Node,
-    retry_policy::{DefaultRetryPolicy, QueryInfo, RetryDecision, RetryPolicy, RetrySession},
-    speculative_execution::SpeculativeExecutionPolicy,
-    Compression,
+use crate::transport::cluster::{Cluster, ClusterData};
+use crate::transport::connection::{
+    BatchResult, Connection, ConnectionConfig, VerifiedKeyspaceName,
 };
+use crate::transport::connection_pool::PoolConfig;
+use crate::transport::iterator::{PreparedIteratorConfig, RowIterator};
+use crate::transport::load_balancing::{
+    LoadBalancingPolicy, RoundRobinPolicy, Statement, TokenAwarePolicy,
+};
+use crate::transport::metrics::Metrics;
+use crate::transport::node::Node;
+use crate::transport::query_result::QueryResult;
+use crate::transport::retry_policy::{
+    DefaultRetryPolicy, QueryInfo, RetryDecision, RetryPolicy, RetrySession,
+};
+use crate::transport::speculative_execution;
+use crate::transport::speculative_execution::SpeculativeExecutionPolicy;
+use crate::transport::Compression;
 use crate::{batch::Batch, statement::StatementConfig};
-use crate::{cql_to_rust::FromRow, transport::speculative_execution};
 
 pub use crate::transport::connection_pool::PoolSize;
 
-use crate::transport::iterator::PreparedIteratorConfig;
 #[cfg(feature = "ssl")]
 use openssl::ssl::SslContext;
 
@@ -98,6 +104,9 @@ pub struct SessionConfig {
     pub disallow_shard_aware_port: bool,
 
     pub default_consistency: Consistency,
+
+    /// If true, full schema is fetched with every metadata refresh.
+    pub fetch_schema_metadata: bool,
     /*
     These configuration options will be added in the future:
 
@@ -142,7 +151,8 @@ impl SessionConfig {
             connect_timeout: std::time::Duration::from_secs(5),
             connection_pool_size: Default::default(),
             disallow_shard_aware_port: false,
-            default_consistency: Consistency::Quorum,
+            default_consistency: Consistency::LocalQuorum,
+            fetch_schema_metadata: true,
         }
     }
 
@@ -253,7 +263,7 @@ impl IntoTypedRows for Vec<result::Row> {
     }
 }
 
-/// Iterator over rows parsed as the given type  
+/// Iterator over rows parsed as the given type\
 /// Returned by `rows.into_typed::<(...)>()`
 pub struct TypedRowIter<RowT: FromRow> {
     row_iter: std::vec::IntoIter<result::Row>,
@@ -317,7 +327,12 @@ impl Session {
 
         node_addresses.extend(resolved);
 
-        let cluster = Cluster::new(&node_addresses, config.get_pool_config()).await?;
+        let cluster = Cluster::new(
+            &node_addresses,
+            config.get_pool_config(),
+            config.fetch_schema_metadata,
+        )
+        .await?;
 
         let session = Session {
             cluster,
@@ -338,7 +353,7 @@ impl Session {
         Ok(session)
     }
 
-    /// Sends a query to the database and receives a response.  
+    /// Sends a query to the database and receives a response.\
     /// Returns only a single page of results, to receive multiple pages use [query_iter](Session::query_iter)
     ///
     /// This is the easiest way to make a query, but performance is worse than that of prepared queries.
@@ -377,7 +392,7 @@ impl Session {
     ///
     /// if let Some(rows) = rows_opt {
     ///     for row in rows.into_typed::<(i32, String)>() {
-    ///         // Parse row as int and text   
+    ///         // Parse row as int and text \
     ///         let (int_val, text_val): (i32, String) = row?;
     ///     }
     /// }
@@ -447,10 +462,10 @@ impl Session {
         Ok(())
     }
 
-    /// Run a simple query with paging  
-    /// This method will query all pages of the result  
+    /// Run a simple query with paging\
+    /// This method will query all pages of the result\
     ///
-    /// Returns an async iterator (stream) over all received rows  
+    /// Returns an async iterator (stream) over all received rows\
     /// Page size can be specified in the [Query](crate::query::Query) passed to the function
     ///
     /// See [the book](https://cvybhu.github.io/scyllabook/queries/paged.html) for more information
@@ -493,7 +508,7 @@ impl Session {
             None => self.retry_policy.new_session(),
         };
 
-        Ok(RowIterator::new_for_query(
+        RowIterator::new_for_query(
             query,
             serialized_values.into_owned(),
             self.default_consistency,
@@ -501,7 +516,8 @@ impl Session {
             self.load_balancer.clone(),
             self.cluster.get_data(),
             self.metrics.clone(),
-        ))
+        )
+        .await
     }
 
     /// Prepares a statement on the server side and returns a prepared statement,
@@ -511,7 +527,7 @@ impl Session {
     /// * Database doesn't need to parse the query
     /// * They are properly load balanced using token aware routing
     ///
-    /// > ***Warning***  
+    /// > ***Warning***\
     /// > For token/shard aware load balancing to work properly, all partition key values
     /// > must be sent as bound values
     /// > (see [performance section](https://cvybhu.github.io/scyllabook/queries/prepared.html#performance))
@@ -583,14 +599,14 @@ impl Session {
     }
 
     /// Execute a prepared query. Requires a [PreparedStatement](crate::prepared_statement::PreparedStatement)
-    /// generated using [`Session::prepare`](Session::prepare)  
+    /// generated using [`Session::prepare`](Session::prepare)\
     /// Returns only a single page of results, to receive multiple pages use [execute_iter](Session::execute_iter)
     ///
     /// Prepared queries are much faster than simple queries:
     /// * Database doesn't need to parse the query
     /// * They are properly load balanced using token aware routing
     ///
-    /// > ***Warning***  
+    /// > ***Warning***\
     /// > For token/shard aware load balancing to work properly, all partition key values
     /// > must be sent as bound values
     /// > (see [performance section](https://cvybhu.github.io/scyllabook/queries/prepared.html#performance))
@@ -667,10 +683,10 @@ impl Session {
         response.into_query_result()
     }
 
-    /// Run a prepared query with paging  
-    /// This method will query all pages of the result  
+    /// Run a prepared query with paging\
+    /// This method will query all pages of the result\
     ///
-    /// Returns an async iterator (stream) over all received rows  
+    /// Returns an async iterator (stream) over all received rows\
     /// Page size can be specified in the [PreparedStatement](crate::prepared_statement::PreparedStatement)
     /// passed to the function
     ///
@@ -723,22 +739,21 @@ impl Session {
             None => self.retry_policy.new_session(),
         };
 
-        Ok(RowIterator::new_for_prepared_statement(
-            PreparedIteratorConfig {
-                prepared,
-                values: serialized_values.into_owned(),
-                default_consistency: self.default_consistency,
-                token,
-                retry_session,
-                load_balancer: self.load_balancer.clone(),
-                cluster_data: self.cluster.get_data(),
-                metrics: self.metrics.clone(),
-            },
-        ))
+        RowIterator::new_for_prepared_statement(PreparedIteratorConfig {
+            prepared,
+            values: serialized_values.into_owned(),
+            default_consistency: self.default_consistency,
+            token,
+            retry_session,
+            load_balancer: self.load_balancer.clone(),
+            cluster_data: self.cluster.get_data(),
+            metrics: self.metrics.clone(),
+        })
+        .await
     }
 
-    /// Perform a batch query  
-    /// Batch contains many `simple` or `prepared` queries which are executed at once  
+    /// Perform a batch query\
+    /// Batch contains many `simple` or `prepared` queries which are executed at once\
     /// Batch doesn't return any rows
     ///
     /// Batch values must contain values for each of the queries
@@ -793,13 +808,13 @@ impl Session {
         .await
     }
 
-    /// Sends `USE <keyspace_name>` request on all connections  
-    /// This allows to write `SELECT * FROM table` instead of `SELECT * FROM keyspace.table`  
+    /// Sends `USE <keyspace_name>` request on all connections\
+    /// This allows to write `SELECT * FROM table` instead of `SELECT * FROM keyspace.table`\
     ///
     /// Note that even failed `use_keyspace` can change currently used keyspace - the request is sent on all connections and
     /// can overwrite previously used keyspace.
     ///
-    /// Call only one `use_keyspace` at a time.  
+    /// Call only one `use_keyspace` at a time.\
     /// Trying to do two `use_keyspace` requests simultaneously with different names
     /// can end with some connections using one keyspace and the rest using the other.
     ///
@@ -844,22 +859,29 @@ impl Session {
         Ok(())
     }
 
-    /// Manually trigger a topology refresh  
-    /// The driver will fetch current nodes in the cluster and update its topology information
+    /// Manually trigger a metadata refresh\
+    /// The driver will fetch current nodes in the cluster and update its metadata
     ///
     /// Normally this is not needed,
-    /// the driver should automatically detect all topology changes in the cluster
-    pub async fn refresh_topology(&self) -> Result<(), QueryError> {
-        self.cluster.refresh_topology().await
+    /// the driver should automatically detect all metadata changes in the cluster
+    pub async fn refresh_metadata(&self) -> Result<(), QueryError> {
+        self.cluster.refresh_metadata().await
     }
 
-    /// Access metrics collected by the driver  
+    /// Access metrics collected by the driver\
     /// Driver collects various metrics like number of queries or query latencies.
     /// They can be read using this method
     pub fn get_metrics(&self) -> Arc<Metrics> {
         self.metrics.clone()
     }
 
+    /// Access cluster data collected by the driver\
+    /// Driver collects various information about network topology or schema.
+    /// They can be read using this method
+    pub fn get_cluster_data(&self) -> Arc<ClusterData> {
+        self.cluster.get_data()
+    }
+
     /// Get [`TracingInfo`] of a traced query performed earlier
     ///
     /// See [the book](https://cvybhu.github.io/scyllabook/tracing/tracing.html)
@@ -869,7 +891,7 @@ impl Session {
             .await
     }
 
-    /// Queries tracing info with custom retry settings.  
+    /// Queries tracing info with custom retry settings.\
     /// Tracing info might not be available immediately on queried node -
     /// that's why the driver performs a few attempts with sleeps in between.
     /// [`GetTracingConfig`] allows to specify a custom querying strategy.
@@ -1113,7 +1135,9 @@ impl Session {
                 let query_info = QueryInfo {
                     error: last_error.as_ref().unwrap(),
                     is_idempotent,
-                    consistency: consistency.unwrap_or(self.default_consistency),
+                    consistency: LegacyConsistency::Regular(
+                        consistency.unwrap_or(self.default_consistency),
+                    ),
                 };
 
                 match retry_session.decide_should_retry(query_info) {
